package com.smith9.geiger.client

import com.smith9.geiger.client.beans.Duration
import org.apache.commons.httpclient.methods.PostMethod
import org.apache.commons.httpclient.HttpClient

import org.joda.time.DateTime

import org.slf4j.Logger
import org.slf4j.LoggerFactory
import com.smith9.geiger.client.beans.Stat

import org.apache.commons.httpclient.methods.multipart.ByteArrayPartSource
import org.apache.commons.httpclient.methods.multipart.FilePart
import org.apache.commons.httpclient.methods.multipart.Part
import org.apache.commons.httpclient.methods.multipart.MultipartRequestEntity
import org.apache.commons.httpclient.HttpStatus

import org.apache.commons.io.FilenameUtils
import groovy.util.slurpersupport.NoChildren

/**
 * Created: Mar 17, 2011
 * @author david
 */

class GeigerClient implements LoaderClient {
    private static final Logger LOG = LoggerFactory.getLogger(GeigerClient.class)
    private URL url
    private long totalSent
    private long totalDuration
//    private long batchesSent
    private HttpClient client
    Stat lastStatCommitted = null

    public GeigerClient(URL url) {
        this.url = url
        client = new HttpClient()
//        batchesSent = 0
    }

    static boolean hasMissingBaseLine(String xml) {
        def root = new XmlSlurper().parseText(xml)
        def found = root.stats.children().find {
            println it
            it.children().size() == 2
        }
        return !(found instanceof NoChildren)
    }

    static Double getPercentDiffFromBaseLine(int percentile, String xml) {
        def root = new XmlSlurper().parseText(xml)
        String percentileAttr = root.percentDiffs.percentiles.@"p_$percentile".toString()
        if (percentileAttr.size() > 0) {
            return percentileAttr.toDouble()
        }
        return null
    }

    String executeQuery(DateTime start, DateTime end, String environment, String productName, String queryGroup, String queryName) {
        PostMethod post = createPost(start, end, productName, environment, queryGroup, queryName)
        return executeQuery(post, queryName)
    }

    String executeQueryAndSave(DateTime start, DateTime end, String environment, String productName, String queryGroup, String queryName, String saveName, String saveGroup) {
        PostMethod post = createPost(start, end, productName, environment, queryGroup, queryName)
        post.addParameter("saveName", saveName)
        post.addParameter("saveGroup", saveGroup)
        return executeQuery(post, queryName)
    }

    private String executeQuery(PostMethod post, String queryName) {
        String xml
        try {
            int statusCode = client.executeMethod(post)
            if (statusCode != HttpStatus.SC_OK) {
                String statusMessage = post.getResponseBodyAsString()
                LOG.error("Error executing query to $url query to post was: $queryName")
                LOG.error("status code " + post.getStatusCode())
                LOG.error(statusMessage)
            } else {
                xml = post.getResponseBodyAsString()
                println xml
            }
        } finally {
            // cleanup - see http://hc.apache.org/httpclient-3.x/tutorial.html
            post.releaseConnection()
        }
        return xml
    }

    private PostMethod createPost(DateTime start, DateTime end, String productName, String environment, String queryGroup, String queryName) {
        PostMethod post = new PostMethod(url.toString() + "/services")
        post.addParameter("fromDate", start.toString())
        post.addParameter("toDate", end.toString())
        post.addParameter("productName", productName)
        post.addParameter("environment", environment)
        post.addParameter("queryGroup", queryGroup)
        post.addParameter("queryName", queryName)
        return post
    }

    String uploadCount(File file) {
        load(file, "counts", FilenameUtils.getExtension(file.name))
    }

    String uploadDuration(File file) {
        load(file, "duration", FilenameUtils.getExtension(file.name))
    }

    private def load(File file, String loadMethod, String fileType) {
        FilePart part = new FilePart("stat.$fileType", file)
        upload(loadMethod, part)
    }

    Stat send(List<Stat> stats) {
        long start = System.currentTimeMillis()
        LoaderConverter converter = createConverter(stats)
        uploadStats(converter, stats)
        logTimeTaken(System.currentTimeMillis() - start, stats.size())
        lastStatCommitted = stats.last()
        return lastStatCommitted
    }

    void logTimeTaken(Long duration, int count) {
        if (LOG.debugEnabled) {
            totalSent += count
            totalDuration += duration
            LOG.debug("Sent $count, 1 took ${duration / count / 1000}s, $totalSent averaged ${totalDuration / totalSent / 1000}s")
        }
    }

    private void uploadStats(LoaderConverter converter, List<Stat> stats) {
        String loadMethod = converter.loadMethod
        String xml = converter.toXml(stats)
        println "xml size = " + xml.length() / (1024 * 1024) / 2 + " mb for ${stats.size()} stats and ${xml.length()} chars"
        String partFilename = "stats.xml"
        FilePart part = new FilePart(partFilename, new ByteArrayPartSource(partFilename, xml.getBytes()))
        upload(loadMethod, part)
    }

    private void upload(String loadMethod, FilePart part) {
        String loadUrl = url.toString() + "/" + loadMethod
        PostMethod post = new PostMethod(loadUrl)
        try {
            List<Parts> parts = [part]
            post.setRequestEntity(new MultipartRequestEntity(parts.toArray(new Part[0]), post.getParams()))
            int statusCode = client.executeMethod(post)
            if (statusCode != HttpStatus.SC_OK) {
                String response = post.getResponseBodyAsString()
                LOG.error("Error uploading stats to $loadUrl")
                LOG.error("status code " + statusCode)
                LOG.error(response)
                throw new RuntimeException("Upload failed, status $statusCode, $response")
                // TODO comment the following out and the above back in, where is logging going to?
                //                println("Error uploading stats to $url xml to post was: $xml")
                //                println("status code " + post.getStatusCode())
                //                println(post.getResponseBodyAsString())
            } else {
                // cleanup - see http://hc.apache.org/httpclient-3.x/tutorial.html
                cleanupResponseBody(post)
            }
        } finally {
            // cleanup - see http://hc.apache.org/httpclient-3.x/tutorial.html
            post.releaseConnection()
        }
    }

    private void cleanupResponseBody(PostMethod post) {
        InputStream response = post.getResponseBodyAsStream()
        BufferedReader bufferedResponse = new BufferedReader(new InputStreamReader(response));
        String line;
        while ((line = bufferedResponse.readLine()) != null) {
            // TODO comment this out
            println line
        }
        bufferedResponse.close();
    }

    private LoaderConverter createConverter(List<Stat> stats) {
        if (stats.first() instanceof Duration) {
            return new DurationLoaderConverter()
        } else {
            return new CountLoaderConverter()
        }
    }
}